import argparse
import json
import logging
import os

from dcm_spec_tools.validator.dicom_file_validator import DicomFileValidator
from dcm_spec_tools.spec_reader.edition_reader import EditionReader


def validate(args, base_path):
    json_path = os.path.join(base_path, 'json')
    with open(os.path.join(json_path, EditionReader.dict_info_json)) as info_file:
        dict_info = json.load(info_file)
    with open(os.path.join(json_path, EditionReader.iod_info_json)) as info_file:
        iod_info = json.load(info_file)
    with open(os.path.join(json_path, EditionReader.module_info_json)) as info_file:
        module_info = json.load(info_file)
    log_level = logging.DEBUG if args.verbose else logging.INFO
    validator = DicomFileValidator(iod_info, module_info, dict_info, log_level)
    error_nr = 0
    for dicom_path in args.dicomfiles:
        error_nr += sum(len(error) for error in list(validator.validate(dicom_path).values()))
    return error_nr


def main():
    parser = argparse.ArgumentParser(
        description='Validates DICOM file IODs')
    parser.add_argument('dicomfiles', help='Path(s) of DICOM files or directories to validate',
                        nargs='+')
    parser.add_argument('--standard-path', '-src',
                        help='Base path with the DICOM specs in docbook and json format',
                        default=os.path.join(os.path.expanduser("~"), 'dcm-spec-tools'))
    parser.add_argument('--revision', '-r',
                        help='Standard revision (e.g. "2014c"), year of '
                             'revision, "current" or "local" (latest '
                             'locally installed)',
                        default='current')
    parser.add_argument('--verbose', '-v', action='store_true',
                        help='Outputs diagnostic information')
    args = parser.parse_args()

    edition_reader = EditionReader(args.standard_path)
    destination = edition_reader.get_revision(args.revision)
    if destination is None:
        print('Failed to get DICOM edition {} - aborting'.format(args.revision))
        return 1

    return validate(args, destination)


if __name__ == '__main__':
    exit(main())



import os


def fixture_path():
    return os.path.join(os.path.dirname(__file__), 'fixtures')


def spec_fixture_path():
    return os.path.join(fixture_path(), 'docbook')


def json_fixture_path():
    return os.path.join(fixture_path(), 'json')



"""
Chapter4Reader collects SOP Class Information information for specific
Storage SOP Classes.
The information is taken from PS3.4 in docbook format as provided by ACR NEMA.
"""
from dcm_spec_tools.spec_reader.spec_reader import (
    SpecReader, SpecReaderLookupError, SpecReaderParseError
)


class Part4Reader(SpecReader):
    """Reads information from PS3.4 in docbook format."""

    def __init__(self, spec_dir):
        super(Part4Reader, self).__init__(spec_dir)
        self.part_nr = 4
        self._sop_class_uids = {}  # SOP Class UID --> chapter
        self._chapters = {}  # chapter --> SOP Class UID list

    def iod_chapter(self, sop_class_uid):
        """Return the chapter in part 3 for the given SOP Class."""
        if not self._sop_class_uids:
            self._read_sop_table('B.5')  # standard SOP Classes
        try:
            return self._sop_class_uids[sop_class_uid]
        except KeyError:
            raise SpecReaderLookupError(
                'SOP Class {} not found'.format(sop_class_uid))

    def iod_chapters(self):
        """Return a dict of the chapter in part 3 for each SOP Class
        listed in table B.5.
        """
        if not self._chapters:
            self._read_sop_table('B.5')
        return self._chapters

    def _read_sop_table(self, chapter):
        table = self._find(self._get_doc_root(),
                           ['chapter[@label="B"]',
                            'section[@label="{}"]'.format(chapter), 'table',
                            'tbody'])
        if table is None:
            raise SpecReaderParseError('SOP Class table in Part 4 not found')
        row_nodes = self._findall(table, ['tr'])
        for row_node in row_nodes:
            column_nodes = self._findall(row_node, ['td'])
            if len(column_nodes) in (3, 4):
                # columns are SOP Class Name, SOP Class UID, IOD Specification
                # and Specialization (only since 2020c)
                uid = self.cleaned_value(self._find_text(column_nodes[1]))
                target_node = self._find(column_nodes[2], ['para', 'olink'])
                if target_node is not None:
                    chapter = target_node.attrib['targetptr'].split('_')[1]
                    self._sop_class_uids[uid] = chapter
                    self._chapters.setdefault(chapter, []).append(uid)
        self._patch_incorrect_values()

    def _patch_incorrect_values(self):
        sc_sop_class_uid = '1.2.840.10008.5.1.4.1.1.7'
        if self._sop_class_uids.get(sc_sop_class_uid, '') == 'A.8':
            self._sop_class_uids[sc_sop_class_uid] = 'A.8.1'
            self._chapters['A.8.1'] = [sc_sop_class_uid]
            del self._chapters['A.8']



"""
Chapter3Reader collects DICOM Information Object Definition information
for specific Storage SOP Classes.
The information is taken from PS3.3 in docbook format as provided by ACR NEMA.
"""
import logging
from itertools import groupby

import sys

from dcm_spec_tools.spec_reader.condition_parser import ConditionParser
from dcm_spec_tools.spec_reader.spec_reader import (
    SpecReader, SpecReaderParseError, SpecReaderLookupError
)


class Part3Reader(SpecReader):
    """Reads information from PS3.3 in docbook format."""

    def __init__(self, spec_dir, dict_info=None):
        super(Part3Reader, self).__init__(spec_dir)
        self.part_nr = 3
        self._condition_parser = None
        self._dict_info = dict_info
        self._iod_descriptions = {}
        self._iod_nodes = {}
        self._module_descriptions = {}
        self._current_refs = []
        self.logger = logging.getLogger()
        if not self.logger.handlers:
            self.logger.addHandler(logging.StreamHandler(sys.stdout))
        if dict_info is not None:
            self._condition_parser = ConditionParser(self._dict_info)

    def iod_description(self, chapter):
        """Return the IOD information for the given chapter.

        The return value is a dict with the entries:
            'title': The display name of the IOD
            'modules': A dictionary of the contained IOD modules with the
                module name as key.
                A module dict value has the following entries:
                'ref': The section in PS3.3 describing the module
                    (e.g. 'C.7.4.2')
                'use': Usage information (e.g. 'M' for mandatory)
        Raises SpecReaderLookupError if the chapter is not found.
        """
        if chapter not in self._iod_descriptions:
            iod_node = self._get_iod_nodes().get(chapter)
            if iod_node:
                description = self._parse_iod_node(iod_node)
                self._iod_descriptions[chapter] = description
        try:
            return self._iod_descriptions[chapter]
        except KeyError:
            raise SpecReaderLookupError(
                'No definition found for chapter {}'.format(chapter))

    def iod_descriptions(self):
        """Return the IOD information dict per chapter.

        The dict has the chapter (e.g. 'A.3') as key and the IOD descriptions
         as value.
        See iod_description() for the format of the IOD descriptions.
        Retired IODs (which have no module list) are omitted.
        """
        return {chapter: self.iod_description(chapter) for chapter in
                self._get_iod_nodes()
                if self.iod_description(chapter)['modules']}

    def module_description(self, section):
        """Return the module information in the given section.

        The return value is a dict with the entries:
            'title': The name of the module
            'attributes': A dictionary of the contained module attributes
                with the tag as key.
                An attribute dict value has the following entries:
                    'name': the tag name
                    'type': the type (1, 1C, 2, 2C, 3)
                    'items': only for sequence tags - contains a dictionary
                        of the module attributes contained in the sequence
        Raises SpecReaderLookupError if the section is not found.
        """
        if section not in self._module_descriptions:
            section_node = self._get_section_node(section)
            if section_node:
                description = self._parse_module_description(section_node)
                self._module_descriptions[section] = description
        try:
            return self._module_descriptions[section]
        except KeyError:
            raise SpecReaderLookupError(
                'No definition found for section {}'.format(section))

    def module_descriptions(self):
        """Return the module attribute information for all IODs.

        The return value is a dict with the section name as key and
        a description dict as value.
        See module_description() for the content of the value dict.
        """
        # ensure that all module attributes are read
        self.iod_descriptions()
        return self._module_descriptions

    def _get_iod_nodes(self):
        if not self._iod_nodes:
            chapter_a = self._find(self._get_doc_root(),
                                   ['chapter[@label="A"]'])
            if chapter_a is None:
                raise SpecReaderParseError('Chapter A in Part 3 not found')
            # ignore A.1
            all_iod_nodes = self._findall(chapter_a, ['section'])[1:]
            iod_def_endings = (
                ' IOD',
                ' Information Object Definition',
                ' Information Objection Definition'  # account for known typo
            )

            iod_sub_nodes = []
            nodes_with_subnodes = []
            for iod_node in all_iod_nodes:
                sub_nodes = self._find_sections_with_title_endings(
                    iod_node, iod_def_endings)
                if sub_nodes:
                    nodes_with_subnodes.append(iod_node)
                    iod_sub_nodes.extend(sub_nodes)
            all_iod_nodes = [node for node in all_iod_nodes if
                             node not in nodes_with_subnodes]
            all_iod_nodes.extend(iod_sub_nodes)
            self._iod_nodes = {node.attrib['label']: node for node in
                               all_iod_nodes}
        return self._iod_nodes

    def _get_section_node(self, section):
        section_parts = section.split('.')
        section_name = section_parts[0]
        search_path = ['chapter[@label="{}"]'.format(section_name)]
        for section_part in section_parts[1:]:
            section_name = section_name + '.' + section_part
            search_path.append('section[@label="{}"]'.format(section_name))
        return self._find(self._get_doc_root(), search_path)

    def _parse_iod_node(self, iod_node):
        return {'title': self._find(iod_node, ['title']).text,
                'modules': self._get_iod_modules(iod_node)}

    def _parse_module_description(self, parent_node):
        table_node = self._find(parent_node, ['table'])
        # handle the case that the parent node is the table itself
        if table_node is None:
            table_node = parent_node
        table_body_node = self._find(table_node, ['tbody'])
        if table_body_node is None:
            return
        rows = self._findall(table_body_node, ['tr'])
        current_level = 0
        current_descriptions = [{}]
        last_tag_id = None
        for row in rows:
            columns = self._findall(row, ['td'])
            tag_name, current_level = self._get_tag_name_and_level(
                columns[0], current_descriptions, current_level, last_tag_id)
            if len(columns) == 4:
                last_tag_id = self._handle_regular_attribute(
                    columns, current_descriptions, last_tag_id, tag_name)
            elif tag_name.startswith('Include'):
                self._handle_included_attributes(columns, current_descriptions)
            else:
                # todo: other entries
                pass
        return current_descriptions[0]

    def _handle_included_attributes(self, columns, current_descriptions):
        include_node = self._find(columns[0], ['para', 'emphasis', 'xref'])
        if include_node is None:
            # todo: functional group macros or similar
            return
        include_ref = include_node.attrib['linkend']
        if self._current_refs and include_ref == self._current_refs[-1]:
            self.logger.debug('Self reference in %s  - ignoring.', include_ref)
            return
        self._current_refs.append(include_ref)
        element, label = self._get_ref_element_and_label(include_ref)
        if label not in self._module_descriptions:
            ref_node = self._get_ref_node(element, label)
            if ref_node is None:
                raise SpecReaderLookupError(
                    'Failed to lookup include reference ' + include_ref)
            # it is allowed to have no attributes (example: Raw Data)
            ref_description = self._parse_module_description(ref_node) or {}
            self._module_descriptions[label] = ref_description
        current_descriptions[-1].setdefault('include', []).append(label)
        self._current_refs.pop()

    def _handle_regular_attribute(self, columns, current_descriptions,
                                  last_tag_id, tag_name):
        tag_id = self._find_text(columns[1])
        tag_type = self._find_text(columns[2])
        if tag_id:
            current_descriptions[-1][tag_id] = {
                'name': tag_name,
                'type': tag_type,
            }
            if self._condition_parser and tag_type in ('1C', '2C'):
                # cond = self._find_all_text(columns[3])
                # index = cond.find('Required if ')
                # if index >= 0:
                #     current_descriptions[-1][tag_id]['desc'] = cond[index:]
                current_descriptions[-1][tag_id][
                    'cond'] = self._condition_parser.parse(
                    self._find_all_text(columns[3]))

            last_tag_id = tag_id
        return last_tag_id

    def _get_ref_node(self, element, label):
        return self._get_doc_tree().find(
            './/{}{}[@label="{}"]'.format(self.docbook_ns, element, label))

    @staticmethod
    def _get_ref_element_and_label(ref):
        element, label = ref.split('_')
        if element == 'sect':
            element = 'section'
        return element, label

    def _get_tag_name_and_level(self, column, current_descriptions,
                                current_level, last_tag_id):
        tag_name = self._find_text(column)
        if not tag_name:
            return '', 0
        start_chars = next(groupby(tag_name))
        level = len(list(start_chars[1])) if start_chars[0] == '>' else 0
        tag_name = tag_name[level:]
        if level > current_level:
            sequence_description = {}
            try:
                current_descriptions[-1][last_tag_id][
                    'items'] = sequence_description
                current_descriptions.append(sequence_description)
            except KeyError:
                # silently ignore error in older specs
                pass
        elif level < current_level:
            current_descriptions.pop()
        return tag_name, level

    def _get_iod_modules(self, iod_node):
        module_table_sections = self._find_sections_with_title_endings(
            iod_node, (' Module Table', ' IOD Modules'))
        if not module_table_sections:
            module_table_sections = self._find_sections_with_title_endings(
                iod_node, ('IOD Entity-Relationship Model',))
        modules = {}
        if len(module_table_sections) == 1:
            module_rows = self._findall(module_table_sections[0],
                                        ['table', 'tbody', 'tr'])
            row_span = 0
            for row in module_rows:
                columns = self._findall(row, ['td'])
                name_index = 0 if row_span > 0 else 1
                if row_span == 0:
                    if 'rowspan' in columns[0].attrib:
                        row_span = int(columns[0].attrib['rowspan'])
                    else:
                        row_span = 1
                name = self._find_text(columns[name_index])
                modules[name] = {}
                try:
                    ref_section = self._find(columns[name_index + 1],
                                             ['para', 'xref']).attrib[
                        'linkend'].split('_')[1]
                except AttributeError:
                    try:
                        ref_section = self._find(
                            columns[name_index + 1], ['xref']).attrib[
                            'linkend'].split('_')[1]
                    except AttributeError:
                        self.logger.warning(
                            'Failed to read module table for %s', name)
                        continue
                modules[name]['ref'] = ref_section
                # make sure the module description is loaded
                self.module_description(ref_section)
                modules[name]['use'] = self._find_text(columns[name_index + 2])
                if (self._condition_parser is not None and
                        modules[name]['use'].startswith('C - ')):
                    modules[name]['cond'] = self._condition_parser.parse(
                        modules[name]['use'])
                else:
                    modules[name]['use'] = modules[name]['use'][0]
                row_span -= 1
        return modules

    def _find_sections_with_title_endings(self, node, title_endings):
        section_nodes = self._findall(node, ['section'])
        found_nodes = []
        for sections_node in section_nodes:
            title_node = self._find(sections_node, ['title'])
            if title_node is not None:
                title = title_node.text
                if any([title.endswith(title_ending) for title_ending in
                        title_endings]):
                    found_nodes.append(sections_node)
        return found_nodes



import json

from dcm_spec_tools.spec_reader.condition import Condition


class DefinitionEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Condition):
            return obj.dict()
        return json.JSONEncoder.default(self, obj)



import logging
import os
import time

import pyfakefs.fake_filesystem_unittest

from dcm_spec_tools import __version__
from dcm_spec_tools.spec_reader.edition_reader import EditionReader


class MemoryEditionReader(EditionReader):
    """Mock class that gets the file contents in constructor instead of downloading them.
    We test this class to avoid real download connections during the test.
    """

    def __init__(self, path, contents=''):
        super(MemoryEditionReader, self).__init__(path=path)
        self.html_contents = contents

    def retrieve(self, html_path):
        with open(html_path, 'w') as html_file:
            html_file.write(self.html_contents)


class EditionReaderTest(pyfakefs.fake_filesystem_unittest.TestCase):
    def setUp(self):
        super(EditionReaderTest, self).setUp()
        self.setUpPyfakefs()
        self.base_path = os.path.join('user', 'dcm-spec-tools')
        self.fs.create_dir(self.base_path)
        logging.disable(logging.CRITICAL)

    def test_empty_html(self):
        reader = MemoryEditionReader(self.base_path, '')
        self.assertIsNone(reader.get_editions())
        self.assertFalse(os.path.exists(os.path.join(self.base_path, reader.json_filename)))

    def test_no_html(self):
        reader = MemoryEditionReader(self.base_path, 'Not html')
        self.assertIsNone(reader.get_editions())
        self.assertFalse(os.path.exists(os.path.join(self.base_path, reader.json_filename)))

    def test_no_editions(self):
        reader = MemoryEditionReader(self.base_path, '<html><A HREF="/medical/dicom/2014a/">test</A><html>')
        self.assertIsNone(reader.get_editions())
        self.assertFalse(os.path.exists(os.path.join(self.base_path, reader.json_filename)))

    def test_valid_editions(self):
        reader = MemoryEditionReader(self.base_path, '<html><A HREF="/bla/">2014a</A>'
                                                     '2014b'
                                                     '<a ref="foo">2015</a>'
                                                     '<a ref="foo">2017e</a>')
        self.assertEqual(['2014a', '2017e'], reader.get_editions())
        self.assertTrue(os.path.exists(os.path.join(self.base_path, reader.json_filename)))

    def test_keep_old_version(self):
        json_path = os.path.join(self.base_path, EditionReader.json_filename)
        self.fs.create_file(json_path, contents='["2014a", "2014c"]')
        file_time = time.time() - 29 * 24 * 60 * 60.0
        os.utime(json_path, (file_time, file_time))
        reader = MemoryEditionReader(self.base_path, '<html><A HREF="/bla/">2018a</A>')
        self.assertEqual(['2014a', '2014c'], reader.get_editions())

    def test_replace_old_version(self):
        json_path = os.path.join(self.base_path, EditionReader.json_filename)
        self.fs.create_file(json_path, contents='["2014a", "2014c"]')
        file_time = time.time() - 31 * 24 * 60 * 60.0
        os.utime(json_path, (file_time, file_time))
        reader = MemoryEditionReader(self.base_path, '<html><A HREF="/bla/">2018a</A>')
        self.assertEqual(['2018a'], reader.get_editions())

    def test_keep_local_version(self):
        json_path = os.path.join(self.base_path, EditionReader.json_filename)
        self.fs.create_file(json_path, contents='["2014a", "2014c"]')
        file_time = time.time() - 31 * 24 * 60 * 60.0
        os.utime(json_path, (file_time, file_time))
        reader = MemoryEditionReader(self.base_path, '<html><A HREF="/bla/">2018a</A>')
        self.assertEqual(['2014a', '2014c'], reader.get_editions(update=False))

    def test_update_if_no_local_version_exists(self):
        json_path = os.path.join(self.base_path, EditionReader.json_filename)
        self.fs.create_file(json_path, contents='[]')
        file_time = time.time() - 31 * 24 * 60 * 60.0
        os.utime(json_path, (file_time, file_time))
        reader = MemoryEditionReader(self.base_path, '<html><A HREF="/bla/">2018a</A>')
        self.assertEqual(['2018a'], reader.get_editions(update=False))

    def test_get_existing_revision(self):
        reader = MemoryEditionReader(self.base_path, '<html><A HREF="/bla/">2014a</A>'
                                                     '<a ref="foo">2014e</a>')
        self.assertEqual('2014a', reader.get_edition('2014a'))

    def test_non_existing_revision(self):
        reader = MemoryEditionReader(self.base_path, '<html><A HREF="/bla/">2014a</A>'
                                                     '<a ref="foo">2014e</a>')
        self.assertIsNone(reader.get_edition('2015a'))

    def test_last_revision_in_year(self):
        reader = MemoryEditionReader(self.base_path, '<html><A HREF="/bla/">2014a</A>'
                                                     '<a ref="foo">2014c</a>'
                                                     '<a ref="foo">2015e</a>')
        self.assertEqual('2014c', reader.get_edition('2014'))

    def test_current_revision(self):
        reader = MemoryEditionReader(self.base_path, '<html><A HREF="/bla/">2014a</A>'
                                                     '<a ref="foo">2014c</a>'
                                                     '<a ref="foo">2015e</a>')
        self.assertEqual('2015e', reader.get_edition('current'))

    def test_check_none_revision(self):
        reader = MemoryEditionReader('/foo/bar', '')
        revision, path = reader.check_revision('none')
        self.assertIsNone(revision)
        self.assertEqual('/foo/bar', path)

    def test_check_revision_existing(self):
        base_path = 'base'
        reader = MemoryEditionReader(base_path, '')
        json_path = os.path.join(base_path, EditionReader.json_filename)
        self.fs.create_file(json_path, contents='["2014a", "2014c", "2015a"]')
        revision, path = reader.check_revision('2014')
        self.assertEqual('2014c', revision)
        self.assertEqual(os.path.join(base_path, '2014c'), path)

    def test_check_revision_nonexisting(self):
        base_path = '/foo/bar'
        reader = MemoryEditionReader(base_path, '')
        json_path = os.path.join(base_path, EditionReader.json_filename)
        self.fs.create_file(json_path, contents='["2014a", "2014c", "2015a"]')
        revision, path = reader.check_revision('2016')
        self.assertIsNone(revision)
        self.assertIsNone(path)

    def test_is_current(self):
        reader = MemoryEditionReader(self.base_path, '<html>'
                                                     '<a ref="foo">2014a</a>'
                                                     '<a ref="foo">2014c</a>'
                                                     '<a ref="foo">2015a</a>'
                                                     '<a ref="foo">2015e</a>')
        self.assertTrue(reader.is_current('2015e'))
        self.assertTrue(reader.is_current('2015'))
        self.assertFalse(reader.is_current('2015a'))
        self.assertFalse(reader.is_current('2015f'))
        self.assertFalse(reader.is_current('2014'))
        self.assertFalse(reader.is_current('2016'))
        self.assertTrue(reader.is_current('current'))
        self.assertTrue(reader.is_current(None))

    def test_is_current_version(self):
        json_path = os.path.join(self.base_path, EditionReader.json_filename)
        self.assertFalse(EditionReader.is_current_version(json_path))
        version_path = os.path.join(json_path, 'version')
        self.fs.create_file(version_path, contents='0.2.1')
        self.assertFalse(EditionReader.is_current_version(json_path))
        os.remove(version_path)
        self.fs.create_file(version_path, contents=__version__)
        self.assertTrue(EditionReader.is_current_version(json_path))

    def test_write_current_version(self):
        json_path = os.path.join(self.base_path, EditionReader.json_filename)
        self.fs.create_dir(json_path)
        self.assertFalse(EditionReader.is_current_version(json_path))
        EditionReader.write_current_version(json_path)
        self.assertTrue(EditionReader.is_current_version(json_path))

    def test_recreate_json_if_needed(self):
        self.create_json_files_called = 0

        def create_json_files(cls, docbook_path, json_path):
            self.create_json_files_called += 1
            for name in ('dict_info.json', 'iod_info.json',
                         'module_info.json', 'uid_info.json'):
                path = os.path.join(json_path, name)
                if not os.path.exists(path):
                    self.fs.create_file(path)

        docbook_path = os.path.join(self.base_path, '2014a', 'docbook')
        for chapter_name in ('part03.xml', 'part04.xml', 'part06.xml'):
            self.fs.create_file(os.path.join(docbook_path, chapter_name))
        orig_create_json_files = MemoryEditionReader.create_json_files
        try:
            MemoryEditionReader.create_json_files = create_json_files
            reader = MemoryEditionReader(self.base_path, '')
            json_path = os.path.join(self.base_path,
                                     EditionReader.json_filename)
            self.fs.create_file(json_path, contents='["2014a", "2014c", "2015a"]')
            reader.get_revision("2014a")
            self.assertEqual(1, self.create_json_files_called)
            reader.get_revision("2014a")
            self.assertEqual(2, self.create_json_files_called)
            json_path = os.path.join(self.base_path, '2014a', 'json')
            EditionReader.write_current_version(json_path)
            reader.get_revision("2014a")
            self.assertEqual(2, self.create_json_files_called)
        finally:
            MemoryEditionReader.create_json_files = orig_create_json_files


"""
Dumps tag information from a DICOM file using information in PS3.6.
"""

import argparse
import json
import os
import re

from pydicom import filereader
from pydicom.errors import InvalidDicomError

from dcm_spec_tools.spec_reader.edition_reader import EditionReader


class DataElementDumper(object):
    tag_regex = re.compile(
        r'(\(?[\dabcdefABCDEF]{4}), *([\dabcdefABCDEF]{4})\)?')

    def __init__(self, dict_info, uid_info, max_value_len, show_image_data,
                 tags):
        self.dict_info = dict_info
        self.max_value_len = max_value_len
        self.level = 0
        self.show_image_data = show_image_data

        self.uid_info = {}
        for uid_dict in uid_info.values():
            self.uid_info.update(uid_dict)

        tags = tags or []
        self.tags = []
        for tag in tags:
            match = self.tag_regex.match(tag)
            if match:
                self.tags.append(
                    '({},{})'.format(match.group(1), match.group(2)))
            else:
                matching = [tag_id for tag_id in dict_info
                            if
                            dict_info[tag_id]['name'].replace(" ", "") == tag]
                if matching:
                    self.tags.append(matching[0])
                else:
                    print('{} is not a valid tag expression - '
                          'ignoring'.format(tag))

    def print_dataset(self, dataset):
        dataset.walk(
            lambda data_set, data_elem: self.print_dataelement(data_set,
                                                               data_elem))

    def print_element(self, tag_id, name, vr, prop, value):
        if self.tags and tag_id not in self.tags:
            return False
        vm = 1 if value else 0
        if isinstance(value, list):
            vm = len(value)
            value = '\\'.join([str(element) for element in value])
        if isinstance(value, bytes):
            value = str(value)[2:-1]
        if isinstance(value, str) and len(value) > self.max_value_len:
            value = value[:self.max_value_len] + '...'

        indent = 2 * self.level
        format_string = '{{}}{{}} {{:{}}} {{}} {{:4}} {{}} [{{}}]'.format(
            40 - indent)
        print(format_string.format(' ' * indent,
                                   tag_id,
                                   name[:40 - indent],
                                   vr,
                                   vm,
                                   prop,
                                   value))
        return True

    def print_dataelement(self, _, dataelement):
        tag_id = '({:04X},{:04X})'.format(dataelement.tag.group,
                                          dataelement.tag.element)
        description = self.dict_info.get(tag_id)
        if description is None:
            name = '[Unknown]'
            vr = dataelement.VR
            prop = ''
        else:
            vr = description['vr']
            name = description['name']
            prop = description['prop']
        value = dataelement.value
        if vr == 'UI':
            # do not rely on pydicom here - we want to use the
            # currently loaded DICOM spec
            value = repr(value)[1:-1]
            value = self.uid_info.get(value, value)
        if vr == 'SQ':
            if self.print_element(tag_id, name, vr, prop,
                                  'Sequence with {} item(s)'.format(
                                      len(value))):
                self.level += 1
                self.print_sequence(dataelement)
                self.level -= 1
        else:
            self.print_element(tag_id, name, vr, prop, value)

    def print_sequence(self, sequence):
        indent = 2 * self.level
        format_string = '{{}}Item {{:<{}}} [Dataset with {{}} element(s)]'\
            .format(56 - indent)
        for i, dataset in enumerate(sequence):
            print(format_string.format(' ' * indent, i + 1, len(dataset)))
            self.level += 1
            dataset.walk(lambda data_set, data_elem:
                         self.print_dataelement(data_set, data_elem))
            self.level -= 1

    def dump_file(self, file_path):
        try:
            print('\n' + file_path)
            dataset = filereader.read_file(
                file_path, stop_before_pixels=self.show_image_data, force=True)
            self.print_dataset(dataset)
        except (InvalidDicomError, KeyError):
            print(
                u'{} is not a valid DICOM file - skipping.'.format(file_path))

    def dump_directory(self, dir_path):
        for root, _, names in os.walk(dir_path):
            for name in names:
                self.dump_file(os.path.join(root, name))


def main():
    parser = argparse.ArgumentParser(
        description='Dumps DICOM information dictionary from '
                    'DICOM file using PS3.6')
    parser.add_argument('dicomfiles',
                        help='Path(s) of DICOM files or directories to parse',
                        nargs='+')
    parser.add_argument('--standard-path', '-src',
                        help='Path with the DICOM specs in docbook '
                             'and json format',
                        default=os.path.join(os.path.expanduser("~"),
                                             'dcm-spec-tools'))
    parser.add_argument('--revision', '-r',
                        help='Standard revision (e.g. "2014c"), year of '
                             'revision, "current" or "local" (latest '
                             'locally installed)',
                        default='current')
    parser.add_argument('--max-value-len', '-ml',
                        help='Maximum string length of displayed values',
                        type=int,
                        default=80)
    parser.add_argument('--show-tags', '-t',
                        help='Show only output for the searched tags. '
                             'Tags can be in the format ####,#### or as the '
                             'dictionary name (e.g. "PatientName").',
                        nargs='*')
    parser.add_argument('--show-image-data', '-id', action='store_false',
                        help='Also show the image data tag (slower)')
    args = parser.parse_args()

    edition_reader = EditionReader(args.standard_path)
    destination = edition_reader.get_revision(args.revision)
    if destination is None:
        print(
            'Failed to get DICOM edition {} - aborting'.format(args.revision))
        return 1

    json_path = os.path.join(destination, 'json')
    with open(os.path.join(json_path,
                           edition_reader.dict_info_json)) as info_file:
        dict_info = json.load(info_file)
    with open(os.path.join(json_path,
                           edition_reader.uid_info_json)) as info_file:
        uid_info = json.load(info_file)

    dumper = DataElementDumper(dict_info, uid_info, args.max_value_len,
                               args.show_image_data,
                               args.show_tags)
    for dicom_path in args.dicomfiles:
        if not os.path.exists(dicom_path):
            print('\n"%s" does not exist - skipping', dicom_path)
        else:
            if os.path.isdir(dicom_path):
                dumper.dump_directory(dicom_path)
            else:
                dumper.dump_file(dicom_path)

    return 0


if __name__ == '__main__':
    exit(main())



import re
from collections import OrderedDict

from dcm_spec_tools.spec_reader.condition import Condition


class ConditionParser(object):
    """Parses the description of type C modules and type 1C and 2C attributes.
    Creates a Condition object representing the condition(s) in the
    description provided that:
    - the condition is related to the value, absence or presence of one
        or more tags in the data set
    - the condition is related only to the data set itself
    All other conditions (including parsable conditions which reference
    other data sets) are ignored.
    The following operators are used for conditions:
    '+' - required if the given tag is present
    '++' - required if the given tag is present and has a value
    '-' - required if the given tag is absent
    '=' - required if the given tag has one of the given values
    '!=' - required if the given tag does not have one of the given values
    '>' - required if the given tag value is greater than the given value
    '<' - required if the given tag value is less than the given value
    '=>' - required if the given tag points to one of the given tags
    """

    tag_expression = re.compile(
        r'(the value of )?(?P<name>[a-zA-Z1-9 \'\-]+)'
        r'(?P<id>\([\dA-Fa-f]{4},[\dA-Fa-f]{4}\))?(,? Value (?P<index>\d))?$')

    operators = OrderedDict([
        (' is greater than ', '>'),
        (' is present and equals ', '='),
        (' is present with a value of ', '='),
        (' value is ', '='),
        (' has a value of more than ', '>'),
        (' has a value greater than ', '>'),
        (' has a value of ', '='),
        (' = ', '='),
        (' at the image level equals ', '='),
        (' equals other than ', '!='),
        (' equals ', '='),
        (' is other than ', '!='),
        (' is present and the value is ', '='),
        (' is present and has a value of ', '='),
        (' is present and has a value', '++'),
        (' is present', '+'),
        (' is sent', '+'),
        (' is not sent', '-'),
        (' is not present', '-'),
        (' is absent', '-'),
        (' is not equal to ', '!='),
        (' is not ', '!='),
        (' is ', '='),
        (' is: ', '='),
        (' are not present', '-'),
        (' are present', '+'),
        (' points to ', '=>')
    ])

    logical_ops = OrderedDict([
        ('and if', 'and'),
        ('and', 'and'),
        ('or if', 'or'),
        ('or', 'or')
    ])

    def __init__(self, dict_info):
        self._dict_info = dict_info

    def parse(self, condition):
        """Parse the given condition string and return a Condition object
         with the required attributes.
        """
        condition_prefixes = ('required if ', 'shall be present if ')
        for prefix in condition_prefixes:
            index = condition.lower().find(prefix)
            if index >= 0:
                condition = condition[len(prefix) + index:]
                condition = self._fix_condition(condition)
                return self._parse_tag_expressions(condition)
        return Condition(ctype='U')

    def _parse_tag_expression(self, condition):
        operator_text = None
        op_offset = None
        for operator in self.operators:
            offset = condition.find(operator)
            if offset > 0 and (op_offset is None or offset < op_offset):
                op_offset = offset
                operator_text = operator
        if operator_text is None:
            return Condition(ctype='U'), None
        operator = self.operators[operator_text]
        rest = condition[op_offset + len(operator_text):]
        if self.operators[operator_text] in ('=', '!=', '>', '<'):
            values, rest = self._parse_tag_values(rest)
        elif self.operators[operator_text] == '=>':
            value_string, rest = self.extract_value_string(rest)
            tag, _ = self._parse_tag(value_string)
            if tag is None:
                return Condition(ctype='U'), None
            values, rest = [str(self._tag_id(tag))], rest
        else:
            values, rest = None, rest.strip()
        result = self._parse_tags(condition[:op_offset], operator, values)
        if not result:
            return Condition(ctype='U'), None

        result.type = ('MU' if 'may be present otherwise'
                               in condition[op_offset:].lower() else 'MN')
        return result, rest

    def _get_other_condition(self, condition_string):
        other_cond_texts = [
            'may be present otherwise if ', 'may be present if '
        ]
        for condition_marker in other_cond_texts:
            index = condition_string.lower().find(condition_marker)
            if index >= 0:
                return self._parse_tag_expressions(
                    condition_string[index + len(condition_marker):])

    @staticmethod
    def _tag_id(tag_id_string):
        group, element = tag_id_string[1:-1].split(',')
        return (int(group, 16) << 16) + int(element, 16)

    def _parse_tag(self, tag_string):
        match = self.tag_expression.match(tag_string.strip())
        if match:
            value_index = (0 if match.group('index') is None
                           else int(match.group('index')) - 1)
            if match.group('id') is not None:
                return match.group('id'), value_index
            tag_name = match.group('name').strip()
            for tag_id, entry in self._dict_info.items():
                if entry['name'] == tag_name:
                    return tag_id, value_index
        return None, None

    def _parse_tag_values(self, value_string):
        value_string, rest = self.extract_value_string(value_string)
        values = value_string.split(', ')
        tag_values = []
        for value in values:
            if ' or ' in value:
                tag_values.extend(value.split(' or '))
            elif value.startswith('or '):
                tag_values.append(value[3:])
            else:
                tag_values.append(value)
        values = []
        for value in tag_values:
            value = value.strip()
            if value.startswith('"') and value.endswith('"'):
                value = value[1:-1].strip()
            values.append(value)
        return values, rest

    def extract_value_string(self, value_string):
        # remove stuff that breaks parser
        value_string = value_string.replace('(Legacy Converted)', '')
        start_index = 0
        rest = None
        while True:
            end_index = -1
            # todo: handle or
            for end_char in (';', '.', 'and ', ', or ', ' or '):
                char_index = value_string.find(end_char, start_index)
                if end_index < 0 or 0 <= char_index < end_index:
                    end_index = char_index
            apo_index = value_string.find('"', start_index)
            if end_index < 0:
                break
            if 0 <= apo_index < end_index:
                start_index = value_string.find('"', apo_index + 1) + 1
                continue
            if end_index > 0:
                if value_string.find(' or ') in [end_index, end_index + 1]:
                    # differentiate between several values and several
                    # conditions - check if the rest is a condition
                    or_cond = self._parse_tag_expressions(
                        value_string[end_index + 3:])
                    if or_cond.type == 'U':
                        start_index = end_index + 4
                        continue
                rest = value_string[end_index:].strip()
                value_string = value_string[:end_index]
                break
        return value_string, rest

    def _parse_tag_expressions(self, condition, nested=False):
        result, rest = self._parse_tag_expression(condition)
        if rest is not None:
            if rest.startswith(', '):
                rest = rest[2:]
            logical_op = None
            for operator in self.logical_ops:
                if rest.startswith(operator + ' '):
                    logical_op = self.logical_ops[operator]
                    condition = rest[len(operator) + 1:]
                    break
            if logical_op is not None:
                next_result = self._parse_tag_expressions(
                    condition, nested=True)
                if next_result.type != 'U':
                    next_result.type = None
                    new_result = Condition(ctype=result.type)
                    cond_list = self._condition_list(logical_op, new_result)
                    cond_list.append(result)
                    cond_list.append(next_result)
                    result.type = None
                    result = new_result
        if not nested and rest is not None:
            other_cond = self._get_other_condition(rest)
            if other_cond is not None and other_cond.type != 'U':
                result.type = 'MC'
                result.other_condition = other_cond
        return result

    def _parse_tags(self, condition, operator, values):
        # this handles only a few cases that are actually found
        if ', and ' in condition:
            return self._parse_tag_composition(
                condition, operator, values, 'and')
        if ', or ' in condition:
            return self._parse_tag_composition(
                condition, operator, values, 'or')
        if ' and ' in condition:
            return self._parse_multiple_tags(
                condition, operator, values, 'and')
        if ' or ' in condition:
            return self._parse_multiple_tags(
                condition, operator, values, 'or')
        return self._result_from_tag_string(condition, operator, values)

    def _parse_tag_composition(self, condition, operator, values, logical_op):
        split_string = ', {} '.format(logical_op)
        conditions = condition.split(split_string)
        result0 = self._parse_tags(conditions[0], operator, values)
        if not result0:
            result = self._parse_tags(condition.replace(
                split_string, split_string.replace(',', '')), operator, values)
        else:
            result = Condition()
            cond_list = self._condition_list(logical_op, result)
            cond_list.append(result0)
            cond_list.append(self._parse_tags(conditions[1], operator, values))
        return result

    def _parse_multiple_tags(self, condition, operator, values, logical_op):
        condition = condition.replace(' {} '.format(logical_op), ', ')
        result = Condition()
        cond_list = self._condition_list(logical_op, result)
        for tag_string in condition.split(', '):
            tag_result = self._result_from_tag_string(
                tag_string, operator, values)
            if tag_result:
                cond_list.append(tag_result)
        if len(cond_list) > 1:
            return result

    @staticmethod
    def _condition_list(logical_op, result):
        cond_list = (result.and_conditions if logical_op == 'and'
                     else result.or_conditions)
        return cond_list

    def _result_from_tag_string(self, tag_string, operator, values):
        tag, index = self._parse_tag(tag_string)
        if tag is not None:
            result = Condition(tag=tag, index=index, operator=operator)
            if values:
                result.values = values
            return result

    @staticmethod
    def _fix_condition(condition):
        index = condition.lower().find(' may be present otherwise')
        if index > 0:
            if condition[index - 1] == ',':
                condition = condition[:index - 1] + '.' + condition[index:]
            elif condition[index - 1] != '.':
                condition = condition[:index] + '.' + condition[index:]
        return condition



import json
import unittest

import os

from dcm_spec_tools.spec_reader.condition import Condition
from dcm_spec_tools.spec_reader.edition_reader import EditionReader
from dcm_spec_tools.tests.test_utils import json_fixture_path


class ConditionReadTest(unittest.TestCase):
    dict_info = None

    @classmethod
    def setUpClass(cls):
        with open(os.path.join(json_fixture_path(),
                               EditionReader.dict_info_json)) as info_file:
            cls.dict_info = json.load(info_file)

    def check_condition(self, cond_dict, cond_type, index=0, op=None,
                        tag=None, values=None, nr_and_cond=0, nr_or_cond=0):
        condition = Condition.read_condition(cond_dict)
        self.assertEqual(cond_type, condition.type)
        self.check_sub_condition(condition, index, op, tag, values,
                                 nr_and_cond, nr_or_cond)
        return condition

    def check_sub_condition(self, condition, index=0, op=None,
                            tag=None, values=None, nr_and_cond=0,
                            nr_or_cond=0):
        self.assertEqual(index, condition.index)
        self.assertEqual(op, condition.operator)
        self.assertEqual(tag, condition.tag)
        self.assertEqual(values, condition.values)
        self.assertEqual(nr_and_cond, len(condition.and_conditions))
        self.assertEqual(nr_or_cond, len(condition.or_conditions))

    def test_read_type_only(self):
        self.check_condition({ "type": "U" }, 'U')

    def test_eq(self):
        cond_dict = {
            "index": 0,
            "op": "=",
            "tag": "(3004,000A)",
            "type": "MN",
            "values": [
                "BEAM",
                "BEAM_SESSION",
                "CONTROL_POINT"
            ]
        }

        def test_condition():
            return self.check_condition(cond_dict, cond_type='MN', op='=',
                                        tag='(3004,000A)',
                                        values=['BEAM', 'BEAM_SESSION',
                                                'CONTROL_POINT'])

        condition = test_condition()
        cond_dict = condition.dict()
        test_condition()
        self.assertEqual('Dose Summation Type is equal to "BEAM", '
                         '"BEAM_SESSION" or "CONTROL_POINT"',
                         condition.to_string(self.dict_info))

    def test_greater(self):
        cond_dict = {
            "index": 0,
            "op": ">",
            "tag": "(0028,0008)",
            "type": "MN",
            "values": ["1"]
        }

        def test_condition():
            return self.check_condition(cond_dict, cond_type='MN', op='>',
                                        tag='(0028,0008)',
                                        values=['1'])

        condition = test_condition()
        cond_dict = condition.dict()
        test_condition()
        self.assertEqual('Number of Frames is greater than 1',
                         condition.to_string(self.dict_info))

    def test_less(self):
        cond_dict = {
            "index": 0,
            "op": "<",
            "tag": "(0028,0008)",
            "type": "MN",
            "values": ["20"]
        }

        def test_condition():
            return self.check_condition(cond_dict, cond_type='MN', op='<',
                                        tag='(0028,0008)',
                                        values=['20'])

        condition = test_condition()
        cond_dict = condition.dict()
        test_condition()
        self.assertEqual('Number of Frames is less than 20',
                         condition.to_string(self.dict_info))

    def test_points_to(self):
        cond_dict = {
            "index": 0,
            "op": "=>",
            "tag": "(0028,0009)",
            "type": "MN",
            "values": [
              "1577061"
            ]
        }

        def test_condition():
            return self.check_condition(cond_dict, cond_type='MN', op='=>',
                                        tag='(0028,0009)',
                                        values=['1577061'])

        condition = test_condition()
        cond_dict = condition.dict()
        test_condition()
        self.assertEqual('Frame Increment Pointer points to (0018,1065) (Frame Time Vector)',
                         condition.to_string(self.dict_info))

    def test_exists(self):
        cond_dict = {
            "index": 0,
            "op": "+",
            "tag": "(7FE0,0010)",
            "type": "MN"
        }

        def test_condition():
            return self.check_condition(cond_dict, cond_type='MN', op='+',
                                        tag='(7FE0,0010)')

        condition = test_condition()
        cond_dict = condition.dict()
        test_condition()
        self.assertEqual('Pixel Data exists',
                         condition.to_string(self.dict_info))

    def test_and_condition(self):
        cond_dict = {
            "and": [
                {
                    "index": 0,
                    "op": "-",
                    "tag": "(0040,E022)"
                },
                {
                    "index": 1,
                    "op": "+",
                    "tag": "(0040,E023)"
                },
                {
                    "index": 0,
                    "op": "!=",
                    "tag": "(0040,E025)",
                    "values": [ "TEST" ]
                }
            ],
            "type": "MU"
        }

        def test_condition():
            condition = self.check_condition(
                cond_dict, cond_type='MU', nr_and_cond=3)
            self.check_sub_condition(
                condition.and_conditions[0], op='-', tag='(0040,E022)')
            self.check_sub_condition(
                condition.and_conditions[1], op='+', tag='(0040,E023)',
                index=1)
            self.check_sub_condition(
                condition.and_conditions[2], op='!=', tag='(0040,E025)',
                values=['TEST'])
            return condition

        condition = test_condition()
        cond_dict = condition.dict()
        test_condition()
        self.assertEqual('DICOM Media Retrieval Sequence is not present and '
                         'WADO Retrieval Sequence[1] exists and WADO-RS '
                         'Retrieval Sequence is not equal to "TEST"',
                         condition.to_string(self.dict_info))

    def test_or_condition(self):
        cond_dict = {
            "or": [
                {
                    "index": 0,
                    "op": "-",
                    "tag": "(0040,4072)"
                },
                {
                    "index": 0,
                    "op": "-",
                    "tag": "(0040,4074)"
                }
            ],
            "type": "MU"
        }

        def test_condition():
            condition = self.check_condition(
                cond_dict, cond_type='MU', nr_or_cond=2)
            self.check_sub_condition(
                condition.or_conditions[0], op='-', tag='(0040,4072)')
            self.check_sub_condition(
                condition.or_conditions[1], op='-', tag='(0040,4074)')
            return condition

        condition = test_condition()
        cond_dict = condition.dict()
        test_condition()
        self.assertEqual('STOW-RS Storage Sequence is not present or '
                         'XDS Storage Sequence is not present',
                         condition.to_string(self.dict_info))

    def test_other_condition(self):
        cond_dict = {
            "index": 0,
            "op": "=",
            "other_cond": {
              "index": 0,
              "op": "+",
              "tag": "(0072,0704)",
              "type": "MN"
            },
            "tag": "(0072,0704)",
            "type": "MC",
            "values": [
              "PALETTE"
            ]
          }

        def test_condition():
            condition = self.check_condition(
                cond_dict, cond_type='MC', op='=',
                tag='(0072,0704)', values=['PALETTE'])
            self.check_sub_condition(
                condition.other_condition, op='+', tag='(0072,0704)')
            self.assertEqual('MN', condition.other_condition.type)
            return condition

        condition = test_condition()
        cond_dict = condition.dict()
        test_condition()



import json
import logging
import sys

from dcm_spec_tools.spec_reader.condition import Condition
from dcm_spec_tools.tag_tools import tag_name_from_id_string, tag_name_from_id


class InvalidParameterError(Exception):
    pass


class IODValidator(object):
    def __init__(self, dataset, iod_info, module_info, dict_info=None,
                 log_level=logging.INFO):
        self._dataset = dataset
        self._iod_info = iod_info
        self._module_info = module_info
        self._dict_info = dict_info
        self.errors = {}
        self.logger = logging.getLogger()
        self.logger.level = log_level
        if not self.logger.handlers:
            self.logger.addHandler(logging.StreamHandler(sys.stdout))

    def validate(self):
        self.errors = {}
        if 'SOPClassUID' not in self._dataset:
            self.errors['fatal'] = 'Missing SOPClassUID'
        else:
            sop_class_uid = self._dataset.SOPClassUID
            if sop_class_uid not in self._iod_info:
                self.errors['fatal'] = 'Unknown SOPClassUID (probably retired): ' + sop_class_uid
            else:
                self._validate_sop_class(sop_class_uid)
        if 'fatal' in self.errors:
            self.logger.error('%s - aborting', self.errors['fatal'])
        else:
            if self.errors:
                self.logger.info('\nErrors\n======')
                for module_name, errors in self.errors.items():
                    self.logger.warning('Module "{}":'.format(module_name))
                    for error_msg in errors:
                        self.logger.warning(error_msg)
        return self.errors

    def add_errors(self, errors):
        for key, value in errors.items():
            self.errors.setdefault(key, []).extend(value)

    def _validate_sop_class(self, sop_class_uid):
        iod_info = self._iod_info[sop_class_uid]
        self.logger.info('SOP class is "%s" (%s)', sop_class_uid,
                         iod_info['title'])
        self.logger.debug('Checking modules for SOP Class')
        self.logger.debug('------------------------------')
        for module_name, module in iod_info['modules'].items():
            errors = self._validate_module(module, module_name)
            if errors:
                self.errors[module_name] = errors

    def _validate_module(self, module, module_name):
        errors = {}
        usage = module['use']
        module_info = self._get_module_info(module['ref'])
        condition = module['cond'] if 'cond' in module else None

        allowed = True
        if usage == 'M':
            required = True
        elif usage == 'U':
            required = False
        else:
            required, allowed = self._object_is_required_or_allowed(condition)
        self._log_module_required(module_name, required, allowed, condition)

        has_module = self._has_module(module_info)
        if not required and not has_module:
            return errors

        if not allowed and has_module:
            for tag_id_string, attribute in module_info.items():
                tag_id = self._tag_id(tag_id_string)
                if tag_id in self._dataset:
                    message = self._incorrect_tag_message(tag_id, 'not allowed', None)
                    errors.setdefault(message, []).append(tag_id_string)
        else:
            for tag_id_string, attribute in module_info.items():
                result = self._validate_attribute(self._tag_id(tag_id_string), attribute)
                if result is not None:
                    errors.setdefault(result, []).append(tag_id_string)
        return errors

    def _log_module_required(self, module_name, required, allowed,
                             condition_dict):
        msg = 'Module "' + module_name + '" is '
        msg += ('required' if required
                else 'optional' if allowed else 'not allowed')
        if condition_dict:
            condition = Condition.read_condition(condition_dict)
            if condition.type != 'U':
                msg += ' due to condition:\n  '
                msg += condition.to_string(self._dict_info)
        self.logger.debug(msg)

    def _incorrect_tag_message(self, tag_id, error_kind, condition_dict):
        msg = 'Tag {} is {}'.format(tag_name_from_id(tag_id, self._dict_info), error_kind)
        if condition_dict:
            condition = Condition.read_condition(condition_dict)
            if condition.type != 'U':
                msg += ' due to condition:\n  '
                msg += condition.to_string(self._dict_info)
        return msg

    def _validate_attribute(self, tag_id, attribute):
        attribute_type = attribute['type']
        # ignore image data and larger tags for now - we don't read them
        if tag_id >= 0x7FE00010:
            return
        has_tag = tag_id in self._dataset
        value_required = attribute_type in ('1', '1C')
        condition_dict = None
        if attribute_type in ('1', '2'):
            tag_required, tag_allowed = True, True
        elif attribute_type in ('1C', '2C'):
            if 'cond' in attribute:
                condition_dict = attribute['cond']
                tag_required, tag_allowed = self._object_is_required_or_allowed(condition_dict)
            else:
                tag_required, tag_allowed = False, True
        else:
            tag_required, tag_allowed = False, True
        error_kind = None
        if not has_tag and tag_required:
            error_kind = 'missing'
        elif tag_required and value_required and self._dataset[tag_id].value is None:
            error_kind = 'empty'
        elif has_tag and not tag_allowed:
            error_kind = 'not allowed'
        if error_kind is not None:
            msg = self._incorrect_tag_message(tag_id, error_kind,
                                              condition_dict)
            return msg


    def _object_is_required_or_allowed(self, condition):
        if isinstance(condition, str):
            condition = json.loads(condition)
        if condition['type'] == 'U':
            return False, True
        required = self._composite_object_is_required(condition)
        if required:
            return True, True
        allowed = (condition['type'] == 'MU' or condition['type'] == 'MC' and
                   self._composite_object_is_required(condition['other_cond']))
        return False, allowed

    def _composite_object_is_required(self, condition):
        if 'and' in condition:
            required = all(self._composite_object_is_required(cond) for cond in condition['and'])
        elif 'or' in condition:
            required = any(self._composite_object_is_required(cond) for cond in condition['or'])
        else:
            required = self._object_is_required(condition)
        return required

    def _object_is_required(self, condition):
        tag_id = self._tag_id(condition['tag'])
        tag_value = None
        operator = condition['op']
        if operator == '+':
            return tag_id in self._dataset
        elif operator == '-':
            return tag_id not in self._dataset
        elif tag_id in self._dataset:
            tag = self._dataset[tag_id]
            index = condition['index']
            if index > 0:
                if index <= tag.VM:
                    tag_value = tag.value[index - 1]
            elif tag.VM > 1:
                tag_value = tag.value[0]
            else:
                tag_value = tag.value
            if tag_value is None:
                return False
            if operator == '++':
                return True
            return self._tag_matches(tag_value, operator, condition['values'])
        return False

    def _has_module(self, module_info):
        for tag_id_string, attribute in module_info.items():
            if attribute['type'] != '3':
                tag_id = self._tag_id(tag_id_string)
                # crude check - attribute could be also in another module
                if tag_id in self._dataset:
                    return True
        return False

    @staticmethod
    def _tag_id(tag_id_string):
        group, element = tag_id_string[1:-1].split(',')
        # workaround for repeating tags -> special handling needed
        if group.endswith('xx'):
            group = group[:2] + '00'
        return (int(group, 16) << 16) + int(element, 16)

    @staticmethod
    def _tag_matches(tag_value, operator, values):
        values = [type(tag_value)(value) for value in values]
        if operator == '=':
            return tag_value in values
        if operator == '!=':
            return tag_value not in values
        if operator == '>':
            return tag_value > values[0]
        if operator == '<':
            return tag_value < values[0]
        if operator == '=>':
            return tag_value in values
        return False

    def _get_module_info(self, module_ref):
        return self._expanded_module_info(self._module_info[module_ref])

    def _expanded_module_info(self, module_info):
        if 'include' in module_info:
            for ref in module_info['include']:
                module_info.update(self._get_module_info(ref))
            del module_info['include']
        if 'items' in module_info:
            module_info['items'] = self._expanded_module_info(module_info['items'])
        return module_info



import json
import logging
import os

import pyfakefs.fake_filesystem_unittest
from pydicom import write_file
from pydicom.dataset import Dataset, FileDataset

from dcm_spec_tools.spec_reader.edition_reader import EditionReader
from dcm_spec_tools.tests.test_utils import json_fixture_path
from dcm_spec_tools.validator.dicom_file_validator import DicomFileValidator


class DicomFileValidatorTest(pyfakefs.fake_filesystem_unittest.TestCase):
    iod_info = None
    module_info = None

    @classmethod
    def setUpClass(cls):
        with open(os.path.join(json_fixture_path(),
                               EditionReader.iod_info_json)) as info_file:
            cls.iod_info = json.load(info_file)
        with open(os.path.join(json_fixture_path(),
                               EditionReader.module_info_json)) as info_file:
            cls.module_info = json.load(info_file)

    def setUp(self):
        super(DicomFileValidatorTest, self).setUp()
        self.setUpPyfakefs()
        logging.disable(logging.CRITICAL)
        self.validator = DicomFileValidator(self.iod_info, self.module_info)

    @staticmethod
    def create_metadata():
        metadata = Dataset()
        metadata.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.7'
        metadata.MediaStorageSOPInstanceUID = '1.2.3'
        metadata.TransferSyntaxUID = '1.2.840.10008.1.2'
        metadata.ImplementationClassUID = '1.3.6.1.4.1.5962.2'
        return metadata

    def assert_fatal_error(self, filename, error_string):
        error_dict = self.validator.validate(filename)
        self.assertEqual(1, len(error_dict))
        name = list(error_dict.keys())[0]
        self.assertEqual(filename, name)
        errors = error_dict[name]
        self.assertEqual({'fatal': error_string}, errors)

    def test_non_existing_file(self):
        self.assert_fatal_error('non_existing', error_string='File missing')

    def test_invalid_file(self):
        self.fs.create_file('test', contents='invalid')
        self.assert_fatal_error('test', error_string='Invalid DICOM file')

    def test_missing_sop_class(self):
        filename = 'test.dcm'
        file_dataset = FileDataset(filename, Dataset(),
                                   file_meta=self.create_metadata())
        write_file(filename, file_dataset, write_like_original=False)
        self.assert_fatal_error(filename, 'Missing SOPClassUID')

    def test_unknown_sop_class(self):
        dataset = Dataset()
        dataset.SOPClassUID = 'Unknown'
        file_dataset = FileDataset('test', dataset,
                                   file_meta=self.create_metadata())
        write_file('test', file_dataset, write_like_original=False)
        self.assert_fatal_error(
            'test',
            'Unknown SOPClassUID (probably retired): Unknown')

    def test_validate_dir(self):
        self.fs.create_dir(os.path.join('foo', 'bar', 'baz'))
        self.fs.create_dir(os.path.join('foo', 'baz'))
        self.fs.create_file(os.path.join('foo', '1.dcm'))
        self.fs.create_file(os.path.join('foo', 'bar', '2.dcm'))
        self.fs.create_file(os.path.join('foo', 'bar', '3.dcm'))
        self.fs.create_file(os.path.join('foo', 'bar', 'baz', '4.dcm'))
        self.fs.create_file(os.path.join('foo', 'baz', '5.dcm'))
        self.fs.create_file(os.path.join('foo1', '6.dcm'))

        self.assertEqual(5, len(self.validator.validate('foo')))

    def test_non_fatal_errors(self):
        dataset = Dataset()
        dataset.SOPClassUID = '1.2.840.10008.5.1.4.1.1.2'  # CT Image Storage
        file_dataset = FileDataset('test', dataset,
                                   file_meta=self.create_metadata())
        write_file('test', file_dataset, write_like_original=False)
        error_dict = self.validator.validate('test')
        self.assertEqual(1, len(error_dict))
        errors = error_dict['test']
        self.assertNotIn('fatal', errors)
